/* vi:set ai sm

	Call Back system by Ann & Arikawa
				by ann@mrit.co.jp
				   arikawa@mrit.mei.co.jp

	cba-1.0 (Login base version)
	cba-2.0 (PPP support)
	cba-3.0 (Direct connection support)
	cba-4.0 (Simplified for one tty callback)
*/

#include "cba.h"

/*	file descripters		*/
static int fi, fo;

/*	time related variables		*/
static struct timeval start;

/*	identification of this program	*/
static char myname[20],
	    mytty[20];
static int  mypid,
	    appid;

/*	identification of this program	*/
static struct termios termorg;

/*	current modem			*/
static struct mdm *modem = NULL;

/*	tty device name			*/
static char *tty;

/*	recorded into wtmp file or not	*/
static int  logged;

/*	reason to terminate application monitor	*/
static char *reason;

/*	flags for timeouts		*/
int	in_to,
	wait_to,
	app_to,
	trace_to;

/*	CBA version			*/
char	*ver = VERSION;

/*	debug flag			*/
int	trace = 0;

/*	operation from terminal		*/
int	term = 0;

/*	caller's ID if known		*/
char	*caller_id = NULL;

/*	externals			*/
extern char *blindread();

extern int addutmp(),
	   addwtmp(),
	   rmutmp();
extern void dmputmp(),
	    dmpwtmp(),
	    cutwtmp();

extern char *mset(),
	    *mdial(),
	    *mflush(),
	    *mconnect();
extern int  hangup();
extern int  setdtr();
extern int  getcar();
extern void dumpmctrl();

extern int writemsg();

char *osid =
#if   defined(OS_SunOS)
	"SUNOS.";
#elif defined(OS_Solaris)
	"SOLARIS.";
#elif defined(OS_Linux)
	"LINUX.";
#elif defined(OS_SVR4)
	"SVR4.";
#elif defined(OS_BSD)
	"BSD.";
#else
	"unknown"
#endif

char *sysutmp = SYSUTMP,
     *syswtmp = SYSWTMP,
     *utmp = UTMPFILE,
     *wtmp = WTMPFILE,
     *wtmp_tmp = WTMP_TMPFILE;

static void	printfd();
static void	cbaabort_core();


static void
cbaabort_hup()
{
    cbaabort_core(SIGHUP, "SIGHUP");
}

static void
cbaabort_int()
{
    cbaabort_core(SIGINT, "SIGINT");
}

static void
cbaabort_term()
{
    cbaabort_core(SIGTERM, "SIGTERM");
}

#ifdef SIGQUIT
static void
cbaabort_quit()
{
    cbaabort_core(SIGQUIT, "SIGQUIT");
}
#endif

#ifdef SIGILL
static void
cbaabort_ill()
{
    cbaabort_core(SIGILL, "SIGILL");
}
#endif

#ifdef SIGABRT
static void
cbaabort_abrt()
{
    cbaabort_core(SIGABRT, "SIGABRT");
}
#endif

#ifdef SIGSEGV
static void
cbaabort_segv()
{
    cbaabort_core(SIGSEGV, "SIGSEGV");
}
#endif

#ifdef SIGPIPE
static void
cbaabort_pipe()
{
    cbaabort_core(SIGPIPE, "SIGPIPE");
}
#endif

/*
 *
 *	CBA main routine
 *
 */

int
main(ac, av)
    int ac;
    char *av[];
{ 
    char *cp;
    int  i;
    int count;
    int status;
    int cin;
    char *cmsg;
    char *etcdir;
    char dirpath[MAXFPATH];
    char prname[MAXPNAME * 3];

    int monstart;
    int mon_maxuse;
    int mon_monint;

    struct vars env[20], *ep;
    struct sys	sysconf;
    struct dev	devconf;
    struct user phones[MAXPHONES + 1], *up;
    struct mtyp mtypes[MAXMTYPES + 1], *mtp;
    struct stat st;
    struct timeval current;
    struct termios tio;

    static void cbaabort();
    static void cbamaxwait();
    static void cbamon_nothing();
    static void cbamon_kill();
    static void cbamon_int();
    static void cbamon_term();
    static void cbamon_hup();
    static void cbacallin();
    static void cbatraceup();
    static void cbatracedown();
    extern int readphones();
    extern int readconf();
    extern int readappli();
    extern int printamsg();
    extern void expand_env();
    extern void trace_modem();
    extern void dumptio();

    /*
     *	Initialize session
     */
    gettimeofday(&start, NULL);
    srand(start.tv_usec);

    sprintf(myname,"cba%s", ver);
    mypid = (int)getpid();

    if (getuid() < (uid_t)10)
    	term = 0;
    else
    	term = 1;

    logged = 0;
    tty = 0;
    etcdir = NUL;
    mtypes[0].name[0] = NUL;

    fi = 0;
    fo = 1;

    /*
     *	Parse parameters
     */
    for( i = 1; i < ac; i++)
    {
        if (!strcmp(av[i], "help") || !strcmp(av[i], "-h"))
        {
	    fprintf(stderr, "call-back system by ann/arikawa (ver.%s)\n",
	    							ver);
	    fprintf(stderr, " options\n");
	    fprintf(stderr, "    /dev/     : specify controling terminal.\n");
	    fprintf(stderr, "    trace     : trace execution.\n");
	    fprintf(stderr, "    trace=#   : specify trace level.\n");
	    fprintf(stderr, "        1: control,   8: utmp \n");
	    fprintf(stderr, "        2: modem,    16: conf. file \n");
	    fprintf(stderr, "        4: tty\n");
	    fprintf(stderr, "    term  : operation from terminal.\n");
	    fprintf(stderr, "    back  : operation on background.\n");
	    fprintf(stderr, "    try   : identical to term.\n");
	    fprintf(stderr, "    who   : list current users. ('wlast' to dump wtmp file\n");
	    fprintf(stderr, "    last  : list log in last(1) format.\n");
	    fprintf(stderr, "    stat  start end [ user ]\n");
	    fprintf(stderr, "          : print statistics log.\n");
	    fprintf(stderr, "    cutwtmp time\n");
	    fprintf(stderr, "          : discard wtmp before specified time.\n");
	    fprintf(stderr, "         time forms mm/dd,hh:mm or mm/dd\n");

	    fprintf(stderr, "    etc=/usr/.../ \n");
	    fprintf(stderr, "          : directory containing setup files.\n");

	    fprintf(stderr, "    wtmp=file \n");
	    fprintf(stderr, "          : WTMP file.\n");

	    fprintf(stderr, "    ver   : print CBA version.\n");
	    fprintf(stderr, "    help  : show this message.\n");
	    exit(0);
	}
	else if (!strcmp(av[i], "/dev/"))
	{
	    static char devpath[MAXFPATH];

	    i++;

            if (!av[i])
	    {
		fprintf(stdout, "/dev/ device_file\n");
		exit(-1);
	    }
	    sprintf(devpath, "/dev/%s", av[i]);
	    tty = devpath;
	}
	else if (!strncmp(av[i], "/dev", 4))
	    tty = av[i];
	else if (!strcmp(av[i], "ver"))
	{ 
	    fprintf(stdout, "Call back system CBA ");
	    fprintf(stdout, "version %s\n", ver);
	    fprintf(stdout, "Copyright 1994 - 1997 ");
	    fprintf(stdout, "Matsushita Research Institute Tokyo, Inc.\n");
	    fprintf(stdout, "[ MRIT / TechieHills ]\n");
	    exit(0);
	}
	else if (!strcmp(av[i], "trace"))
	    trace = 1;
	else if (!strncmp(av[i], "trace=", 6))
	    trace= atoi(av[i] + 6);

	else if (!strcmp(av[i], "term") || !strcmp(av[i], "try"))
	    term = 1;
	else if (!strcmp(av[i], "back") || !strcmp(av[i], "bg"))
	    term = 0;
	else if (!strncmp(av[i], "etc=", 4))
	    etcdir = av[i] + 4;
	else if (!strncmp(av[i], "wtmp=", 5))
	{
	    wtmp = av[i] + 5;
	    if (!strcmp(wtmp, "-"))
	    	wtmp = NULL;
	}
	else if (!strcmp(av[i], "who"))
	{ 
	    dmputmp(utmp);
	    exit(0);
	}
	else if (!strcmp(av[i], "wlast"))
	{ 
	    dmputmp(wtmp ? wtmp: syswtmp);
	    exit(0);
	}
	else if (!strcmp(av[i], "last"))
	{
	    if (!av[i + 1])
	    {
	    	if (wtmp)
		    execl(LAST, "last(cba)", "-f", wtmp, NULL);
		else
		    execl(LAST, "last(cba)", NULL);
	    }
	    else
	    	if (wtmp)
		    execl(LAST, "last(cba)", "-f", wtmp, av[i + 1], NULL);
		else
		    execl(LAST, "last(cba)", av[i + 1], NULL);

	    exit(0);
	}
	else if (!strcmp(av[i], "log"))
	{ 
	    extern char *msgfile;
	    int fd;
	    char cbuf[2048];

	    fd = open(msgfile, O_RDONLY);
	    if (fd >= 0)
	    {
	    	while((i = read(fd, cbuf, 2048)) > 0)
		    write(1, cbuf, i);
		close(fd);
	    }
	    else
	    	fprintf(stderr, "Cannot open %s.", msgfile);
	    exit(0);
	}
	else if (!strcmp(av[i], "stat"))
	{
	    if (!av[i + 1] || !av[i + 2])
	    {
	    	fprintf(stderr, "specify start and end time (and login name)\n");
	    	exit(-1);
	    }

	    dmpwtmp(wtmp, av[i + 1], av[i + 2], av[i + 3]);
	    exit(0);
	}
	else if (!strcmp(av[i], "userstat"))
	{
	    if (!av[i + 1] || !av[i + 2])
	    {
	    	fprintf(stderr, "specify start and end time (and login name)\n");
	    	exit(-1);
	    }

	    trace |= TRACE_CTRL;
	    dmpwtmp(wtmp, av[i + 1], av[i + 2], av[i + 3]);
	    exit(0);
	}
	else if (!strcmp(av[i], "ttystat"))
	{
	    if (!av[i + 1] || !av[i + 2])
	    {
	    	fprintf(stderr, "specify start and end time (and login name)\n");
	    	exit(-1);
	    }

	    trace |= TRACE_TTY;
	    dmpwtmp(wtmp, av[i + 1], av[i + 2], av[i + 3]);
	    exit(0);
	}
	else if (!strcmp(av[i], "cutwtmp"))
	{
	    if (!av[i + 1])
	    {
	    	fprintf(stderr, "specify time to cut\n");
	    	exit(-1);
	    }
	    cutwtmp(wtmp, wtmp_tmp, av[i + 1]);
	    exit(0);
	}
	else
	{
	    if (!term)
		writemsg(SYSTM, "unknown parameter '%s'.\n", av[i]);
	    fprintf(stderr, "unknown parameter '%s'.\n", av[i]);
	    exit(-1);
	}
    }

    if (trace)
	writemsg(TRACE, "------- %s start -------\n", myname);

    alarm(30);

    /*
     *	open tty
     */
    if (tty)
    {
	fi = fo = open(tty, O_RDWR | O_NDELAY);

	if (fi < 0)
	{
	    writemsg(SYSTM, "cannot open dvice %s. : %s\n",
						tty, strerror(errno));
	    exit(-1);
	}

	/* check modem control line */
	if (trace & TRACE_MODEM)
	    dumpmctrl(fi);
    }

    cp = tty ? tty: ttyname(fi);

    if (cp == NULL)
    {
	writemsg(SYSTM, "couldn't get tty.\n");
	exit(-1);
    }
    strcpy(mytty, cp);
    tcgetattr(fi, &termorg);

    rmutmp(utmp, mytty);

    /*
     *	Read configuration file and initialize ttys
     */
    sprintf(dirpath, "%s/%s", etcdir? etcdir: ETCDIR, CONFTTY);

    if (trace & TRACE_CTRL)
    	writemsg(TRACE, "reading device configuration file %s.\n", dirpath);

    switch(readconf(dirpath, mytty, &sysconf, &devconf, mtypes))
    {
    case 1:
	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s: setup termio structure.\n", mytty);

	tcsetattr(fi, TCSANOW, &(devconf.tio));

	if (trace & TRACE_TTY)
	    dumptio(&(devconf.tio));
	break;

    case 0:
	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s: suppress to setup termio structure.\n", mytty);

	if (tty)
	{
	    writemsg(TRACE, "terminal %s not found in %s.\n", mytty, dirpath);
	    exit(-1);
	}

	break;

    default:
    	exit(-1);

    }


    if (!term)
    {
	if (stat(utmp, &st))
	{
	    writemsg(TRACE, "operating system restarted.\n");
	    addwtmp(wtmp, "~", "reboot", NULL);
	}
	addutmp(utmp, mytty, myname, "starting.");
	addutmp(sysutmp, mytty, myname, "starting.");
    }

    signal(SIGHUP,  cbaabort_hup);
    signal(SIGINT,  cbaabort_int);
    signal(SIGTERM, cbaabort_term);
#ifdef SIGQUIT
    signal(SIGQUIT, cbaabort_quit);
#endif
#ifdef SIGILL
    signal(SIGILL,  cbaabort_ill);
#endif
#ifdef SIGABRT
    signal(SIGABRT, cbaabort_abrt);
#endif
#ifdef SIGSEGV
    signal(SIGSEGV, cbaabort_segv);
#endif
#ifdef SIGPIPE
    signal(SIGPIPE, cbaabort_pipe);
#endif

    /*
     *	Initialize string environment
     */

    ep = env;
    ep->name = '%'; ep->val = "%";		ep++;
    ep->name = '}'; ep->val = "}";		ep++;
    ep->name = 'v'; ep->val = myname;		ep++;
    ep->name = 't'; ep->val = devconf.name;	ep++;
    ep->name = 'b'; ep->val = devconf.baud;	ep++;
    ep->name = 'H'; ep->val = devconf.peer;	ep++;
    ep->name = 'G'; ep->val = devconf.ip;	ep++;
    ep->name = NUL;

    /*
     *	Initialize modem
     */
    if (tty)
    {
    	modem = &devconf.modem;

    	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s: hangup modem.\n", mytty);

	if (hangup(fi, mytty, modem))
	{
	    writemsg(SYSTM, "%s: cannot hangup modem.\n", mytty);
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "%s: setting up modem.\n", mytty);
	    mset(fi, fo, devconf.modem.setup, mytty, NULL);
	    cbaabort();
	}

	trace_modem(fi, fo, mytty, &(devconf.modem), TT_INIT, env);

    	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s: modem initialize.\n", mytty);
	
    	if (!mset(fi, fo, devconf.modem.reset, mytty, NULL)
	 || !mset(fi, fo, devconf.modem.setup, mytty, NULL))
	    cbaabort();
    }

    sprintf(dirpath, "waiting(%5d).", mypid);

    if (!term)
    {
	addutmp(utmp, mytty, myname, dirpath);
	addutmp(sysutmp, mytty, myname, dirpath);
    }

    /*
     *	Waiting call-in
     */
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "waiting call-in.\n");

    wait_to = 0;
    trace_to = 0;
    signal(SIGUSR1, cbatraceup);
    signal(SIGUSR2, cbatracedown);
/*
    signal(SIGHUP,  cbamaxwait);
    signal(SIGTERM, cbamaxwait);
*/
    if (sysconf.maxwait > 0)
    {
	signal(SIGALRM, cbamaxwait);
	alarm(sysconf.maxwait * 60);
    }

    /*
     *	Read connect message
     */
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "%s waiting connect message for %d sec.\n", mytty,
				sysconf.maxwait? sysconf.maxwait * 60: 5);

    caller_id = NULL;
    if (tty)
    {
	setdtr(fi, 1);

	cmsg = mconnect(fi, fo, NULL, mytty, sysconf.maxwait?
		  				sysconf.maxwait * 60: 5,
					     &devconf.modem, env);
	mflush(fi, mytty);
    }
    else
	cmsg = "/dev/(terminal)";

    alarm(0);
    if (wait_to)
    {
	writemsg(TRACE, "%s periodic termination.\n", mytty);
	cbaabort();
    }

    if (!cmsg)
    {
	writemsg(SYSTM, "%s no connect message\n", mytty);
    	cbaabort();
    }

    /*
     *	Read phone number file and process call-in session
     */
    sprintf(dirpath, "%s/%s", etcdir? etcdir: ETCDIR, CONFPHO);

    if (trace & TRACE_CTRL)
    	writemsg(TRACE, "read phone number file %s.\n", dirpath);

    if (readphones(dirpath, phones, MAXPHONES))
    	cbaabort();

    cin = 0;
#ifdef CALLIN
    /*
     *	Check registerd call-in.
     */
    if (caller_id)
    {
	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s call-in connected from %s.\n", mytty, caller_id);

	for(up = phones; up->name[0]; up++)
	    if (!strcmp(up->pid, CALLIN))
	    {
	    	char *cp1,*cp2, c1, c2;

		cp1 = up->phone;
		cp2 = caller_id;

		/* compare those numbers ignoring symbols */
		while(*cp1 && *cp2)
		{
		    c1 = *cp1;
		    if (c1 < '0' || c1 > '9')
		    {
		    	cp1++;
		    	continue;
		    }

		    c2 = *cp2;
		    if (c2 < '0' || c2 > '9')
		    {
		    	cp2++;
		    	continue;
		    }

		    if (c1 != c2)
		    	break;

		    cp1++;
		    cp2++;
		}

		while(*cp1 && (*cp1 < '0' || *cp1 > '9'))
		    cp1 ++;

		while(*cp2 && (*cp2 < '0' || *cp2 > '9'))
		    cp2 ++;

		if (!*cp1 && !*cp2)
		    break;
	    }

	if (up->name[0])
	{
	    writemsg(TRACE, "%s call-in connected(%s:%s): %s.\n",
					mytty, up->name, up->where, cmsg);
	    cin = 1;

	    sprintf(dirpath, "%s(%d).", up->name,mypid);
	    if (!term)
	    {
		addutmp(utmp, mytty, myname, dirpath);
		addutmp(sysutmp, mytty, myname, dirpath);
	    }
	}
    }
#endif /* CALLIN */


    if (!cin)
    {
	/*
	 *	Process call-in session
	 */
	printfd(fo, "\n");
	writemsg(TRACE, "%s call-in connected: %s\n", mytty, cmsg);

	sprintf(dirpath, "call-in(%d).", mypid);
	if (!term)
	{
	    addutmp(utmp, mytty, myname, dirpath);
	    addutmp(sysutmp, mytty, myname, dirpath);
	}

	in_to = 0;
	signal(SIGALRM, cbacallin);
	alarm(sysconf.to_ac);

	lsleep(2);

	/* Access verification */
	printfd(fo, "*********************************\n");

	cp = blindread(fi, fo, "Access verification", 0);

	if (in_to)
	{
	    writemsg(MAINT, "access verification timeout.\n");
	    printfd(fo, "\n** time out **\n");
	    cbaabort();
	}

	if (!cp)
    	    cbaabort();

	if (strcmp(cp, sysconf.verify))
	{
	    *(cp + 10) = NUL;
    	    writemsg(SECUR, "rejected by access verification: %s\n", cp);
    	    cbaabort();
	}

	printfd(fo, "Hello! \n");

	alarm(sysconf.to_in);
	in_to = 0;

	/* call-in session */
	do
	{
	    char phone[MAXPHONEID], name[MAXNAME];

	    /*
	     *	read Phone ID
	     */
	    cp = blindread(fi, fo, "Phone ID", 1);
	    if (in_to)
	    {
		writemsg(SECUR, "Phone ID timeout.\n");
		printfd(fo, "\n** time out **\n");
		cbaabort();
	    }
	    if (cp == NULL)
		cbaabort();
	    *(cp + MAXPHONEID - 1) = NUL;
	    strcpy(phone, cp);

	    /*
	     *	read Login Name
	     */
	    cp = blindread(fi, fo, "   Login", 1);
	    if (in_to)
	    {
		writemsg(SECUR, "login timeout.\n");
		printfd(fo, "\n** time out **\n");
		cbaabort();
	    }
	    if (cp == NULL)
		cbaabort();
	    *(cp + MAXNAME - 1) = NUL;
	    strcpy(name, cp);

	    /*
	     *	search phone number table
	     */
	    for(up = phones; up->name[0]; up++)
		if (!strcmp(phone, up->pid) &&
		    !strcmp(name,  up->name))
		    break;

	    if (up->name[0] == NUL)
	    {
		writemsg(SECUR, "incorrect access %s\n", name);
		up = NULL;
		if (term)
		{
		    printfd(fo, "Unknown user/phone ID pair.\n");
	    	    cbaabort();
		}
		printfd(fo, "Retry. \n");
		continue;
	    }

	    /* search modem type table */
	    for(mtp = mtypes; mtp->name[0]; mtp++)
		if (!strcmp(up->setup, mtp->name))
		    break;

	    if (mtp->name[0] == NUL)
	    {
		writemsg(SYSTM, "'%s' for user %s@%s not found.\n",
			    			up->setup, name, up->where);
		printfd(fo, "System registration error. \n");

		if (!term)
		{
		    up = NULL;
		    printfd(fo, "Retry. \n");
		    continue;
		}
	    }

	} while(up == NULL);
    }

    alarm(0);

    {	/* convert EUC to JIS */
	char *cp, *dp;
	int len;
	int kanji;

	dp = prname;
	cp = up->where;
	len = kanji = 0;
	while(*cp)
	{
	    if (*cp & 0x80)
	    {
		if (!kanji)
		{
		    kanji = 1;
		    len += 3;
		    strcpy(dp, "\033$B");
		    dp += 3;
		}
		len+=2;
		*dp++ = *cp++ & 0x7f;
		*dp++ = *cp++ & 0x7f;
	    }
	    else
	    {
		if (kanji)
		{
		    kanji = 0;
		    len += 3;
		    strcpy(dp, "\033(B");
		    dp += 3;
		}
		len++;
		*dp++ = *cp++;
	    }

	    if (len >= MAXPNAME * 3 - 4)
		break;
	}

	if (kanji)
	{
	    kanji = 0;
	    len += 3;
	    strcpy(dp, "\033(B");
	    dp += 3;
	}

	*dp = NUL;
    }

    ep->name = 'u'; ep->val = up->name;	ep++;
    ep->name = 'p'; ep->val = prname;	ep++;
    ep->name = 'T'; ep->val = up->term;	ep++;
    ep->name = NUL;


#ifdef DIRECT
    if (!strcmp(up->phone, DIRECT))
    {
    	struct passwd *pwp;

	pwp = getpwnam(up->name);
	if (!pwp)
	{
	    writemsg(SYSTM, "User %s not registerd for direct call-in.\n",
    							up->name);
	    printfd(fo, "\n** time out **\n");
	    cbaabort();
	}

	alarm(sysconf.to_in);
    	in_to = 0;

	cp = blindread(fi, fo, "Password", 1);
	alarm(0);
	if (in_to)
	{
	    writemsg(SECUR, "password for direct call-in(%s:%s) timeout.\n",
    							up->name, prname);
	    printfd(fo, "\n** time out **\n");
	    cbaabort();
	}
	if (cp == NULL)
	    cbaabort();

	if (strcmp(pwp->pw_passwd, crypt(cp, pwp->pw_passwd)))
	{
	    writemsg(SECUR, "incorrect password for direct call-in(%s:%s).\n",
    							up->name, prname);
	    cbaabort();
	}

	writemsg(SECUR, "direct call-in for %s(%s:%s) accepted.\n",
    					up->name, prname, up->setup);
    }
    else
#endif
    if (!cin)
    {
	printfd(fo, "Call back to %s from %s by %s soon.\n",
    						prname, mytty, up->setup);
	printfd(fo, "See you later.\n");
	printfd(fo, "[ MRIT / TechieHills ]\n");

	sprintf(dirpath, "accepted(%d).", mypid);
	if (!term)
	{
	    addutmp(utmp, mytty, up->name, dirpath);
	    addutmp(sysutmp, mytty, up->name, dirpath);
	}
	writemsg(TRACE, "call-in for %s(%s:%s) accepted.\n",
    					up->name, prname, up->setup);

	/*
	 *	Hang up
	 */
	if (tty)
	{
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "closing %s.\n", tty);

	    lsleep(1);

	    if (hangup(fi, mytty, modem))
	    {
		writemsg(SYSTM, "%s: cannot hangup modem.\n", mytty);
		cbaabort();
	    }

	    /*
	     * SIGINT when modem send break
	     */
	    devconf.tio.c_iflag |= BRKINT;
	    tcsetattr(fi, TCSANOW, &(devconf.tio));

	    trace_modem(fi, fo, mytty, &(devconf.modem), TT_BACK, env);
	}

	lsleep(5);

	/*
	 *	Call-back
	*/

	for(count = sysconf.retry; count > 0; count--)
	{
	    static char cbuf[] = "dialing(xxx)";

	    char dial[MAXFLINE];
	    char *cp, *dp;

	    /*	Setup user specified function for modem	*/
    	    if (tty)
	    {
		if (!mset(fi, fo, mtp->setup, mytty, NULL))
		    cbaabort();

		setdtr(fi, 1);
	    }

	    /*	Dial	*/
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "dialing to %s:%s.(%d)\n",
						up->name, prname, count);

	    sprintf(cbuf, "dialing(%d)", count);
	    if (!term)
	    {
		addutmp(utmp, mytty, up->name, cbuf);
		addutmp(sysutmp, mytty, up->name, cbuf);
	    }

	    sprintf(dial, devconf.modem.dial, up->phone);
	    /* remove '-', '(' and ')' */
	    for(cp = dp = dial; *cp; cp++)
		if (*cp != '-' && *cp != '(' && *cp != ')')
		    *dp++ = *cp;
	    *dp = NUL;

	    /* dial */
	    if ((cmsg = tty ? mdial(fi, fo, dial, sysconf.to_back, mytty)
			    : "(terminal)"))
		/* connection established */
		break;

	    /* connection not established */
    	    writemsg(MAINT, "dial failed(%s:%s).\n", up->name, prname);

	    /* hangup modem */
	    if (tty && hangup(fi, mytty, modem))
	    {
		writemsg(SYSTM, "%s: cannot hangup modem.\n", mytty);
		cbaabort();
	    }

	    if (count <= 1)
		continue;

	    /* waiting interval */
	    if (tty)
		mset(fi, fo, devconf.modem.hook ,mytty, NULL);

	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "dialing interval %d sec.\n", sysconf.intvl);

	    sleep(sysconf.intvl);

	    if (tty)
	    {
		if (hangup(fi, mytty, modem))
		    writemsg(SYSTM, "%s: cannot hangup modem.\n", mytty);

		mset(fi, fo, devconf.modem.reset ,mytty, NULL);

		trace_modem(fi, fo, mytty, &(devconf.modem), TT_RTRY, env);
	    }

	}

	mflush(fi, mytty);

	if (count <= 0)
	{
	    writemsg(MAINT, "give up to connect %s:%s by %s.\n",
					up->name, prname, up->setup);
	    cbaabort();
	}
    }

    writemsg(TRACE, "%s %s(%s:%s) connected:%s.\n",
		    		mytty, up->name, prname, up->setup, cmsg);
    if (!term)
    {
	addutmp(utmp, mytty, up->name, "connected");
	addutmp(sysutmp, mytty, up->name, "connected");
	logged = 1;
	addwtmp(wtmp, mytty, up->name, up->where);
    }

    /*
     *	Fork
     */
#ifdef OS_Solaris       /* hmmmmmm. why solaris vfork returns 0 for both */
    appid = tty? fork(): 0;
#else
    appid = tty? vfork(): 0;
#endif

    if (appid == -1)
    {
	writemsg(SYSTM, "vfork failed: %s\n", strerror(errno));
	cbaabort();
    }

    if (appid == 0)
    {
	char *apps[MAXAARGV];
	struct hostent *hp;

#ifndef ADDRLEN
# define ADDRLEN 16
#endif
	char myaddr[ADDRLEN], praddr[ADDRLEN];
	struct in_addr in;
	char envt[90];
	int suid;
	int flag;

	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "forked to launch application.\n");

	/*
	 *	Find  applications
	 */
	hp = gethostbyname(devconf.ip);
	if (!hp)
	{
	    writemsg(SYSTM, "host %s unknown.\n", devconf.ip);
	    cbaabort();
	}
	in = *(struct in_addr *)(hp->h_addr_list);
	memcpy(&in, hp->h_addr_list[0], sizeof(in));
	cp = inet_ntoa(in);
	strncpy(myaddr, cp, strlen(cp));

	hp = gethostbyname(devconf.peer);
	if (!hp)
	{
	    writemsg(SYSTM, "host %s unknown.\n", devconf.peer);
	    cbaabort();
	}
	in = *(struct in_addr *)(hp->h_addr_list);
	memcpy(&in, hp->h_addr_list[0], sizeof(in));
	cp = inet_ntoa(in);
	strncpy(praddr, cp, strlen(cp));

	ep->name = 'h'; ep->val = praddr;	ep++;
	ep->name = 'g'; ep->val = myaddr;	ep++;
	ep->name = NUL;

	/* set environment variable TERM and window size */

	sprintf(envt, "TERM=%s", up->term);
	putenv(envt);

#ifdef TIOCSSIZE		/* for SunOS */
	{
	    struct ttysize ts;

	    ts.ts_lines = TERM_ROW;
	    ts.ts_cols =  TERM_COL;
	    ioctl(fi, TIOCSSIZE, &ts);
	}
#elif defined(TIOCSWINSZ)	/* for Solaris */
	{
	    struct winsize w;

	    w.ws_row = TERM_ROW;
	    w.ws_col = TERM_COL;
	    w.ws_xpixel = 
	    w.ws_ypixel = 0;
	    ioctl(fi, TIOCSWINSZ, &w);
	}
#else
	sprintf(envt, "LINES=%s",   TERM_ROW);
	putenv(envt);
	sprintf(envt, "COLUMNS=%s", TERM_COL);
	putenv(envt);
#endif

	/* parse application configuration file */

	sprintf(dirpath, "%s/%s", etcdir? etcdir: ETCDIR, CONFAPP);

	suid = readappli(dirpath, up->appli, apps, MAXAARGV, env);
	if (suid < 0)
	    cbaabort();

	/* print per application message */
	sprintf(dirpath, "%s/%s", etcdir? etcdir: ETCDIR, CONFAMSG);
    	lsleep(1);
	printamsg(fo, dirpath, up->appli, env);
	
	if (trace & TRACE_CTRL)
	    for(i = 0; apps[i]; i++)
		writemsg(TRACE, "%d: '%s'.\n", i, apps[i]);

	/*
	 *	Exec
	 */
	gettimeofday(&current, NULL);
	if (!term)
	{
	    addutmp(utmp, mytty, up->name, up->where);
	    addutmp(sysutmp, mytty, up->name, up->where);
	    addwtmp(syswtmp, mytty, up->name, up->where);
	}

	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "launch application '%s'(%d).\n", apps[0], mypid);

	/*
	 *	set user and group ID
	 */

	if (suid && tty)
	{
	    struct passwd *pwp;

	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "setting user id to %s.\n", up->name);

	    pwp = getpwnam(up->name);
	    if (!pwp)
	    {
		writemsg(MAINT, "unknown user %s.\n", up->name);
		cbaabort();
	    }

	    if (setgid(pwp->pw_gid))
	    {
		writemsg(MAINT, "setgid to %d failed.\n", pwp->pw_gid);
		cbaabort();
	    }

	    if (setuid(pwp->pw_uid))
	    {
		writemsg(MAINT, "setuid to %s failed.\n", up->name);
		cbaabort();
	    }
	}

	if (tty)
	{
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "redirecting stdio.\n");

	    if (fi != 0)
	    {
		close(0);
		if (dup(fi) == -1)
		{
		    writemsg(SYSTM, "stdin redirect failed(%s) : %s\n",
						apps[0], strerror(errno));
		    exit(errno);
		}
	    }

	    if (fo != 1)
	    {
		close(1);
		if (dup(fo) == -1)
		{
		    writemsg(SYSTM, "stdout redirect failed(%s) : %s\n",
						apps[0], strerror(errno));
		    exit(errno);
		}
	    }

	    if (fo != 2)
	    {
		close(2);
		if (dup(fo) == -1)
		{
		    writemsg(SYSTM, "stderr redirect failed(%s) : %s\n",
						apps[0], strerror(errno));
		    exit(errno);
		}
	    }

	    flag = fcntl(0, F_GETFL, 0);
	    fcntl(0, F_SETFL, flag & ~O_NONBLOCK);

	    tcgetattr(0, &tio);
	    tio.c_oflag |=  ONLCR;
	    tio.c_cflag &= ~CLOCAL;
	    tio.c_lflag |= (ICANON | ECHO);
	    tcsetattr(0, TCSANOW, &tio);
	}


	execv(apps[0], &apps[1]);

	writemsg(SYSTM, "exec failed(%s) : %s\n", apps[0], strerror(errno));
	exit(errno);
    }

    /*
     *	Waiting termination of the application. (or timed out)
     */
    app_to = 0;
    reason = NULL;
    signal(SIGHUP,  cbamon_hup);
    signal(SIGINT,  cbamon_int);
    signal(SIGTERM, cbamon_term);

    gettimeofday(&current, NULL);
    monstart = current.tv_sec;

    signal(SIGALRM, cbamon_nothing);
    alarm(0);

    mon_maxuse = sysconf.maxuse * 60;
    mon_monint = sysconf.monint * 60;

    if (trace & TRACE_CTRL)
	writemsg(TRACE, "waiting for termination of %s\n", up->appli);

    do
    {
	int ret;
	int spend;

	/* checking carrier */
	if (tty)
	{
	    int fd;

	    fd = open(tty, O_RDONLY);
	    if (fd >= 0)
	    {
		if (getcar(fd) == 0)
		{
		    writemsg(MAINT, "%s(%s) carrier lost.\n", mytty, up->name);
		    if (!reason)
			reason = "carrier lost";
	    	    cbamon_kill();
		}
		else if (trace & TRACE_CTRL)
		    writemsg(TRACE, "%s carrier O.K.\n", mytty);
		close(fd);
	    }
	}

	/* reset alarm */
	gettimeofday(&current, NULL);
	spend = current.tv_sec - monstart;

	if (app_to == 0)
	{
	    if (mon_maxuse > 0 && mon_monint > 0)
		alarm(mon_maxuse - spend > mon_monint?  mon_monint:
							mon_maxuse - spend);

	    else if (mon_monint > 0)
		alarm(mon_monint);

	    else if (mon_maxuse > 0)
		alarm(mon_maxuse - spend);
	}

	/* waiting user process */
	status = 0;
	ret = waitpid(-1, &status, 0);
	if (ret == -1)
	{
	    if (errno == ECHILD)
	    	break;

	    if (errno != EINTR)
	    {
		writemsg(SYSTM, "%s(%s) waitpid error: %s\n",
	    				mytty, up->name, strerror(errno));
		if (!reason)
		    reason = strerror(errno);
		cbamon_kill();
	    }
	    else if (trace & TRACE_CTRL)
		writemsg(TRACE, "%s(%s) interrupted.\n", mytty, up->name);
	}
	else
	{
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "'%s'(%d) status changed(0x%x).\n",
						up->appli, ret, status);
	    if (WIFEXITED(status))
	    	break;
	}

	/* checking timeout */
	gettimeofday(&current, NULL);
	spend = current.tv_sec - monstart;

	if (mon_maxuse && mon_maxuse <= spend)
	{
	    if (!reason)
		reason = "time exceed";
	    writemsg(MAINT, "%s(%s) time exceed.\n", mytty, up->name);
	    cbamon_kill();
	}
	else if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s(%s) %d sec. lefft.\n", mytty, up->name,
	    					mon_maxuse - spend);

    } while(app_to < 3);


    alarm(0);

    gettimeofday(&current, NULL);
    if (logged)
    {
	addwtmp(wtmp, mytty, NULL, "unknown");
	logged = 0;
    }

    if (reason)
	writemsg(TRACE, "'%s:%d'(%s) terminates by %s.\n",
					up->appli, appid, up->name, reason);
    else
	writemsg(TRACE, "'%s:%d'(%s) terminates.\n",
					up->appli, appid, up->name);

    lsleep(5);


    /*
     *	Hang up
     */
    if (tty)
    {
	tcgetattr(fi, &tio);
	tio.c_cflag |= CLOCAL;
	tio.c_lflag &= ~(ICANON | ECHO);
	tcsetattr(fi, TCSANOW, &tio);

	if (hangup(fi, mytty, modem))
	    writemsg(SYSTM, "%s: cannot hangup modem.\n", mytty);

    	if (ttyname(fo) == NULL)
	    fo = fi;	/* if 'fo' is already closed,
	    		   use fi re-opened in hangup() */

	trace_modem(fi, fo, mytty, &(devconf.modem), TT_DISC, env);

	mset(fi, fo, devconf.modem.reset, mytty, NULL);
    }

    tcsetattr(fi, TCSANOW, &termorg);

    close(fi);
    close(fo);

    rmutmp(utmp, mytty);
    rmutmp(sysutmp, mytty);
    writemsg(TRACE, "%s (%s) disconnected.\n", up->name, mytty);

    return 0;
}

static void
cbaabort_core(sig, signame)
    int  sig;
    char *signame;
{
    signal(sig, NULL);
    writemsg(TRACE, "%s aborting by signal %s.\n", mytty, signame);

    chdir("/usr/tmp");

    rmutmp(utmp, mytty);
    if (logged)
	addwtmp(wtmp, mytty, NULL, "abort");

    if (tty && hangup(fi, mytty, modem))
	    writemsg(SYSTM, "%s: cannot hangup modem.\n", mytty);

    tcsetattr(fi, TCSANOW, &termorg);

    kill(getpid(), sig);
    sleep(5);

    exit(-1);
}

static void
cbaabort()
{
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "aborting.\n");

    rmutmp(utmp, mytty);
    if (logged)
	addwtmp(wtmp, mytty, NULL, "abort");

    if (tty && hangup(fi, mytty, modem))
	    writemsg(SYSTM, "%s: cannot hangup modem.\n", mytty);

    tcsetattr(fi, TCSANOW, &termorg);

    exit(-1);
}

static void
cbamaxwait()
{
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "%s user application interrupted\n", mytty);

    wait_to = 1;
    alarm(0);

    return;
}

static void
cbamon_nothing()
{
    signal(SIGALRM, cbamon_nothing);
    return;
}

static void
cbamon_kill()
{
    char cbuf[10];

    if (trace & TRACE_CTRL)
	writemsg(TRACE, "%s killing session.\n", mytty);

    signal(SIGALRM, cbamon_kill);
    sprintf(cbuf, "pid%d", appid);

    switch(app_to)
    {
    case 0:
    	if (!term)
	    addutmp(utmp, mytty, cbuf, "SIGHUP");
	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s sending SIGHUP to %d.\n", mytty, appid);
	if (kill(appid, SIGHUP))
	    app_to = 3;
	else
	    alarm(3);
	break;

    case 1:
    	if (!term)
	    addutmp(utmp, mytty, cbuf, "SIGTERM");
	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "%s sending SIGTERM to %d.\n", mytty, appid);
	if (kill(appid, SIGTERM))
	    app_to = 3;
	else
	    alarm(2);
	break;

    case 2:
    	if (!term)
	    addutmp(utmp, mytty, cbuf, "SIGKILL");
	writemsg(MAINT, "%s sending SIGKILL to %d.\n", mytty, appid);
	kill(appid, SIGKILL);
    }
    app_to ++;
    return;
}


static void
cbamon_hup()
{
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "session interrupted by SIGHUP\n");
    if (!reason)
	reason = "hangup";
    cbamon_kill();
}

static void
cbamon_int()
{
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "session interrupted by SIGHUP\n");
    if (!reason)
	reason = "interrupt";
    cbamon_kill();
}

static void
cbamon_term()
{
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "session interrupted by SIGTERM\n");
    if (!reason)
	reason = "termination signal";
    cbamon_kill();
}

static void
cbacallin()
{
    if (trace & TRACE_CTRL)
	writemsg(TRACE, "call-in session interrupted\n");

    in_to = 1;
    alarm(0);
    return;
}

static void
cbatraceup()
{
    if (trace >= 0)
	trace ++;
    writemsg(TRACE, "signaled to increase trace level(-> 0x%x).\n", trace);

    trace_to = 1;
    return;
}

static void
cbatracedown()
{
    if (trace > 0)
	trace --;
    writemsg(TRACE, "signaled to decrease trace level(-> 0x%x).\n", trace);

    trace_to = 1;
    return;
}

static void
printfd(fd, form, p1, p2, p3, p4, p5, p6, p7)
    int fd;
    char *form;
    char *p1, *p2, *p3, *p4, *p5, *p6, *p7; 
{
    char cbuf[MAXFLINE];

    sprintf(cbuf, form, p1, p2, p3, p4, p5, p6, p7);
    write(fd, cbuf, strlen(cbuf));
    tcdrain(fd);
}

