/* vi:set ai sm

	modem utility routines for CBA
				by ann@mrit.co.jp

*/

#include "cba.h"

#define MMWAIT		5	/* sec   */
#define SETUP_RETRY	3	/* times */
#define SETUP_WAIT	2	/* sec   */
#define CONNECT_RETRY	10	/* times */
#define CONNECT_SESSION	60	/* sec   */
#define BREAKTIME	0	/* time to send zero-valued bits for break */

#define RESPONSE_OK		{"OK", NULL}
#define RESPONSE_ERROR		{"ERROR", NULL}

#define RESPONSE_CONNECT	{"CONNECT", NULL}
#define RESPONSE_RING		{"RING", "RRING", NULL}
#define RESPONSE_BUSY		{"BUSY", "NO ANSWER", "NO CARRIER", "NOT READY", NULL}
#define RESPONSE_DELAYED	{"DELAYED", NULL}

#define RESPONSE_CALLING	{"CALLING-ID=", NULL} /* Digital Celler & PHS */
#define RESPONSE_CONNECTED	{"CONNECTED-ID=", NULL}	 /* PHS */
#define RESPONSE_OUTOFSERV	{"OUT OF SERVICE", NULL} /* PHS */

static char *r_ok[]	= RESPONSE_OK;
static char *r_error[]	= RESPONSE_ERROR;
static char *r_connect[]= RESPONSE_CONNECT;
static char *r_ring[]	= RESPONSE_RING;
static char *r_busy[]	= RESPONSE_BUSY;
static char *r_delayed[]= RESPONSE_DELAYED;
static char *r_calling[]= RESPONSE_CALLING;
static char *r_cncted[] = RESPONSE_CONNECTED;
static char *r_ooserv[] = RESPONSE_OUTOFSERV;
static char mcr[]   = "\r";

extern int trace;
extern int wait_to, trace_to;

extern writemsg();
void dumpmctrl();
int getcar();
int setdtr();
int waitfd();

extern char *caller_id;

/* wait t/10 sec. */
#ifdef NO_USLEEP
lsleep(t)
    int t;
{
    int fds[2];
    struct timeval to;

    pipe(fds);
    to.tv_sec = 0;
    to.tv_usec= t * 100000;

    waitfd(fds[0], &to);

    close(fds[0]);
    close(fds[1]);
}
#endif


/*
 *	read a line from Modem
 *		timeout when wait second later
 */

static char *
mgets(fi, wait)
    int fi;
    int wait;
{
    static char cbuf[MAXFLINE];
    int  nread;
    char *cp;
    int  len;
    int  spend, spendm;

    struct timeval ts, te;
    struct timeval to;

    if (!isatty(fi))
    {
	writemsg(SYSTM, "logical tty is not phisical tty.\n");
    	return NULL;
    }

    cp = cbuf;
    len =
    nread =
    spend =
    spendm = 0;

    gettimeofday(&ts, NULL);

    while(wait > spend && len < MAXFLINE)
    {
	char c;
	int r;

	to.tv_sec  = wait - spend;
	to.tv_usec = 0;
	if (spendm != 0)
	{
	    to.tv_sec -= 1;
	    to.tv_usec = (1000 - spendm) * 1000;
	}

	r = waitfd(fi, &to);
	if (r < 0)
	{
	    if (!wait_to)
		writemsg(SYSTM, "select/poll failed: %s\n", strerror(errno));
	    return NULL;
	}

	if (r == 0)
	    nread = 0;
	else if ((nread = read(fi, &c, 1)) == 0)
	{
	    writemsg(SYSTM, "inconsistent select/poll - read result\n");
	    return NULL;
	}

	/* error occur */
	if (nread < 0)
	{
	    writemsg(SYSTM, "read failed: %s\n", strerror(errno));
	    return NULL;
	}

	/* compute time spend */
	gettimeofday(&te, NULL);
	spend  = te.tv_sec  - ts.tv_sec;
	spendm = (te.tv_usec - ts.tv_usec) / 1024;

	if (spendm < 0)
	{
	    spend -= 1;
	    spendm = 1000 + spendm;
	}

	/* no character read */
	if(nread == 0)
	    continue;

	/* suppress newline */
	if (c == '\n' || c == '\r' || c == '\012')
	{
	    if (cp != cbuf)
		break;
	    else
	        continue;
	}

	*cp++ = c;
	len ++;
    }

    *cp = NUL;

    if (trace & TRACE_MODEM)
    {
    	if (nread > 0)
	    writemsg(TRACE, "(mgets) read: '%s'\n", cbuf);

	else if (wait > 0)
	    writemsg(TRACE, "(mgets) timeout\n");
    }

    return nread > 0 ? cbuf : NULL;
}

mflush(fd, tty)
    int  fd;
    char *tty;
{
    struct termios term, termorg;
    char *resp;

    tcgetattr(fd, &termorg);
    memcpy(&term, &termorg, sizeof(struct termios));

    term.c_lflag    &= ~ICANON;

    term.c_cc[VTIME] = 0;
    term.c_cc[VMIN]  = 1;

    tcsetattr(fd, TCSANOW, &term);

    /* flush data not yet read/write in this host */
    tcflush(fd, TCIOFLUSH);

    lsleep(1);
    /* discard all characters modem keeps */
    while((resp = mgets(fd, 0)))
	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "%s(mflush) ignore: '%s'\n", tty, resp);

    tcsetattr(fd, TCSANOW, &termorg);
}

char *
mset(fi, fo, setup, tty, etitle)
    int fi, fo;
    char *setup;
    char *tty;
    char *etitle;
{
    char *resp;
    char **rcp;
    char *ftty;
    int  count;
    int  slen;

    struct termios term, termorg;

    tcgetattr(fi, &termorg);
    memcpy(&term, &termorg, sizeof(struct termios));

    ftty = NULL;
    if (*tty == '/')
    	ftty = strchr(tty + 1, '/') + 1;
    if (!ftty)
    	ftty = tty;

    term.c_lflag    &= ~ICANON;

    term.c_cc[VTIME] = 0;
    term.c_cc[VMIN]  = 1;

    tcsetattr(fi, TCSANOW, &term);

    slen = strlen(setup);

    for(count = SETUP_RETRY; count > 0; count--)
    {
    	/* flush data not yet read/write in this host */
	tcflush(fi, TCIFLUSH);
	tcflush(fo, TCOFLUSH);

    	/* discard all characters modem keeps */
	while((resp = mgets(fi, 0)))
	    if (trace & TRACE_MODEM)
		writemsg(TRACE, "%s(mset) ignore: '%s'\n", ftty, resp);

	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "%s(mset) write %d: '%s'\n", ftty, slen, setup);

    	/* send string to modem */
	if (write(fo, setup, slen) < 0)
	{
	    writemsg(SYSTM, "(mset) write failed: %s\n", strerror(errno));
	    tcsetattr(fi, TCSANOW, &termorg);
	    return NULL;
	}
	write(fo, mcr, sizeof(mcr) - 1);
	tcdrain(fo);

	do
	{
    	    /* get modem response */
	    resp = mgets(fi, SETUP_WAIT);

	    /* no response */
	    if (!resp)
	    {
		if (trace & (TRACE_MODEM | TRACE_CTRL))
		    writemsg(TRACE, "%s(mset) modem %s(%d) no response.\n",
		    					ftty, setup, count);
		break;
	    }

	    /* ignore newline and modem echo */
	    if (!*resp || !strncmp(resp, setup, slen))
	    	continue;

	    /* response OK */
	    for(rcp = r_ok; *rcp; rcp++)
	    	if (!strncmp(resp, *rcp, strlen(*rcp)))
		{
		    if (trace & TRACE_MODEM)
			writemsg(TRACE, "%s(mset) %s(%d) response %s(OK).\n",
						ftty, setup, count, resp);
		    tcsetattr(fi, TCSANOW, &termorg);
		    return resp;
		}

	    /* response BUSY */
	    for(rcp = r_busy; *rcp; rcp++)
	    	if (!strncmp(resp, *rcp, strlen(*rcp)))
		{
		    if (trace & TRACE_MODEM)
			writemsg(TRACE, "%s(mset) %s(%d) response %s(BUSY).\n",
						ftty, setup, count, resp);
		    tcsetattr(fi, TCSANOW, &termorg);
		    return resp;
		}

	    /* retry on error */
	    for(rcp = r_error; *rcp; rcp++)
	    	if (!strncmp(resp, *rcp, strlen(*rcp)))
		{
		    if (trace & TRACE_MODEM)
			writemsg(TRACE, "%s(mset) modem %s(%d) error: %s.\n",
						ftty, setup, count, resp);
		    continue;
		}

	    if (etitle)
		writemsg(TRACE, "%s %s: %s\n", ftty, etitle, resp);
	    else
		writemsg(MAINT, "%s(mset) modem unknown response (%s): '%s'\n",
		    					ftty, setup, resp);

	} while(1);

    }


    writemsg(MAINT, "%s modem %s failure.\n", ftty, setup);

    tcsetattr(fi, TCSANOW, &termorg);
    return NULL;
}


char *
mconnect(fi, fo, ret, tty, wait, modem, env)
    int  fi, fo;
    char *tty;
    char **ret;
    int  wait;
    struct mdm *modem;
    struct vars *env;
{
    char *resp;
    char **rcp;
    char *ftty;
    int  count;

    char *cp;
    int  i;

    static char cid[MAXFLINE];
    static char cbuf[MAXFLINE];

    struct termios term, termorg;
    struct timeval t, t2;

    cid[0] = NUL;

    ftty = NULL;
    if (*tty == '/')
    	ftty = strchr(tty + 1, '/') + 1;
    if (!ftty)
    	ftty = tty;

    tcgetattr(fi, &termorg);
    memcpy(&term, &termorg, sizeof(struct termios));

    term.c_lflag    &= ~ICANON;

    term.c_cc[VTIME] = 0;
    term.c_cc[VMIN]  = 1;

    tcsetattr(fi, TCSANOW, &term);

    gettimeofday(&t, NULL);

    for(count = CONNECT_RETRY; count > 0; count--)
    {
	if (!(resp = mgets(fi, wait)))
	{
	    if (trace & TRACE_MODEM)
		writemsg(TRACE, "%s(mconnect) modem connect timeout.\n", ftty);
	    tcsetattr(fi, TCSANOW, &termorg);
	    return NULL;
	}

	gettimeofday(&t2, NULL);
	if (t2.tv_sec - t.tv_sec < CONNECT_SESSION)
	    count = CONNECT_RETRY;
	t.tv_sec = t2.tv_sec;

	/* ignore newline */
	if (!*resp)
	{
	    if (trace & TRACE_MODEM)
		writemsg(TRACE, "%s(mconnect) ignore empty line.\n", ftty);
	    continue;
	}

	/* response returned str */
	if (ret)
	    for(rcp = ret; *rcp; rcp++)
		if (!strncmp(resp, *rcp, strlen(*rcp)))
		{
		    if (trace & TRACE_MODEM)
			writemsg(TRACE, "%s(mconnect:%d) response %s.\n",
		    					ftty, count, resp);
		    tcsetattr(fi, TCSANOW, &termorg);
		    return resp;
		}

	/* response Calling ID */
	for(rcp = r_calling; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		if (trace & TRACE_MODEM)
		    writemsg(TRACE, "%s caller : %s.\n", ftty, resp);
		strcpy(cid, resp + strlen(*rcp));
		break;
	    }
	if (*rcp)
	    continue;

	/* response Connected ID */
	for(rcp = r_cncted; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		if (trace & TRACE_MODEM)
		    writemsg(TRACE, "%s callee : %s.\n", ftty, resp);
		break;
	    }
	if (*rcp)
	    continue;

	/* response Out of Service */
	for(rcp = r_ooserv; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		if (trace & TRACE_MODEM)
		    writemsg(TRACE, "%s(mconnect) out of service.\n", ftty);
		break;
	    }
	if (*rcp)
	    continue;

	/* response Ring */
	for(rcp = r_ring; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		void trace_modem();

		if (trace & TRACE_MODEM)
		    writemsg(TRACE, "%s(mconnect:%d) ignore ring (%s).\n",
		    					ftty, count, resp);

		cp = resp + strlen(*rcp);
		while( *cp == ' ' || *cp == '\t' || *cp == '/')
		    cp++;

		if (*cp >= '0' && *cp <= '9')
		{
		    if (trace & TRACE_MODEM)
			writemsg(TRACE, "%s caller : %s.\n", ftty, cp);
		    strcpy(cid, cp);
		}

		if (modem)
		    trace_modem(fi, fo, ftty, modem, TT_RING, env);

		break;
	    }
	if (*rcp)
	    continue;

	/* response Busy */
	for(rcp = r_busy; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		if (trace & TRACE_MODEM)
		    writemsg(TRACE, "%s(mconnect:%d) ignore busy (%s).\n",
		    					ftty, count, resp);
		break;
	    }
	if (*rcp)
	    continue;

	/* response Connect */
	for(rcp = r_connect; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		if (trace & TRACE_MODEM)
		    writemsg(TRACE, "%s(mconnect) modem connected.\n", ftty);
		tcsetattr(fi, TCSANOW, &termorg);

		if (cid)
		{
		    strcpy(cbuf, resp);

		    if (strlen(cbuf) >= MAXFLINE - 20)
		    	cbuf[MAXFLINE - 20] = NUL;

		    strcpy(cbuf + strlen(cbuf), " ");

		    if (strlen(cbuf) + strlen(cid) >= MAXFLINE)
		    	cid[MAXFLINE - strlen(cbuf) - 1] = NUL;

		    strcpy(cbuf + strlen(cbuf), cid);
		    caller_id = cid;
		    return cbuf;
		}
		else
		    return resp;
	    }

	/* unknown response */
	writemsg(MAINT, "%s(mconnect:%d) modem unknown response: '%s'\n",
		    					ftty, count, resp);
    }

    writemsg(SYSTM, "%s modem connect failed.\n", ftty);
    tcsetattr(fi, TCSANOW, &termorg);
    return NULL;
}

char *
mdial(fi, fo, dial, wait, tty)
    int fi, fo;
    char *dial;
    char *tty;

    int wait;
{
    char *resp;
    char *ftty;
    int  dlen;

    char *rchar[(sizeof(r_busy) + sizeof(r_delayed)) / sizeof(char *) + 2];
    char **rcp, **ccp;

    struct termios term, termorg;

    ftty = NULL;
    if (*tty == '/')
    	ftty = strchr(tty + 1, '/') + 1;
    if (!ftty)
    	ftty = tty;

    tcgetattr(fi, &termorg);
    memcpy(&term, &termorg, sizeof(struct termios));

    term.c_lflag    &= ~ICANON;

    term.c_cc[VTIME] = 0;
    term.c_cc[VMIN]  = 1;

    tcsetattr(fi, TCSANOW, &term);


    dlen = strlen(dial);
    rcp = rchar;
    *rcp++ = dial;

    for(ccp = r_busy; *ccp; ccp++)
    	*rcp++ = *ccp;
    for(ccp = r_delayed; *ccp; ccp++)
    	*rcp++ = *ccp;

    *rcp = NULL;

    /* flush data not yet read/write in this host */
    tcflush(fi, TCIFLUSH);
    tcflush(fo, TCOFLUSH);

    if (trace & TRACE_MODEM)
	writemsg(TRACE, "%s(mdial) write: '%s'\n", ftty, dial);

    /* send string to modem */
    if (write(fo, dial, dlen) < 0)
    {
	writemsg(SYSTM, "(mdial) write failed: %s\n", strerror(errno));
	tcsetattr(fi, TCSANOW, &termorg);
	return NULL;
    }
    write(fo, mcr, sizeof(mcr) - 1);
    tcdrain(fo);

    do
    {
	/* get modem response */
	resp = mconnect(fi, fo, rchar, ftty, wait, NULL, NULL);

	/* error */
	if (!resp)
	{
	    tcsetattr(fi, TCSANOW, &termorg);
	    return NULL;
	}

	/* ignore newline and modem echo */
	if (!*resp || !strncmp(resp, dial, dlen))
	    continue;

	/* response BUSY */
	for(rcp = r_busy; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		writemsg(TRACE, "%s(mdial) modem busy: '%s'\n", ftty, resp);
		tcsetattr(fi, TCSANOW, &termorg);
		return NULL;
	    }

	/* response DELAYED */
	for(rcp = r_delayed; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		writemsg(TRACE, "%s(mdial) modem delayed: '%s'\n", ftty, resp);
		tcsetattr(fi, TCSANOW, &termorg);
		return NULL;
	    }

	/* retry on error */
	for(rcp = r_error; *rcp; rcp++)
	    if (!strncmp(resp, *rcp, strlen(*rcp)))
	    {
		if (trace & TRACE_MODEM)
		    writemsg(SYSTM, "%s(mdial) modem error.\n", ftty);
		continue;
	    }

	/* connect */
	tcsetattr(fi, TCSANOW, &termorg);
	return resp;

    } while(1);

    /* never executed */
    tcsetattr(fi, TCSANOW, &termorg);
    return NULL;
}

int
hangup(fd, tty, mp)
    int fd;
    char *tty;
    struct mdm *mp;
{
    char *resp;

    struct termios term;
    struct timeval time;
    int cflag;
    int st;
    speed_t ospeed;
    speed_t ispeed;

    char *tname;

    tname = ttyname(fd);

    if (tname == NULL)
    {
	int newfd;

	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "(hangup) terminal %s is closed.\n", tty);
	tname = tty;
	newfd = open(tty, O_RDWR | O_NDELAY);
	if (newfd < 0)
	    return -1;

	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "(hangup) terminal %s is re-opened.\n", tty);

	if (fd != newfd)
	{
	    dup2(newfd, fd);
	    if (trace & TRACE_MODEM)
		writemsg(TRACE, "(hangup) terminal descripter duplicated.\n");
	    close(newfd);
	}
    }

    tcgetattr(fd, &term);
    cflag = term.c_cflag;
    ospeed = cfgetospeed(&term);
    ispeed = cfgetispeed(&term);

    if (trace & TRACE_MODEM)
	writemsg(TRACE, "(hangup) ispeed = 0x%x, ospeed = 0x%x.\n",
							ispeed, ospeed);

    /*
     * ignore modem break
     */
    term.c_iflag |=  IGNBRK;
    term.c_iflag &= ~BRKINT;
    term.c_lflag &= ~ICANON;
    term.c_cflag |=  CLOCAL;

    term.c_cc[VTIME] = 0;
    term.c_cc[VMIN]  = 1;

#ifdef RPORT_RESET
    term.c_cflag &= ~CRTSCTS;

    /*
     *  set baud rate to 9600 bps
     */
    cfsetispeed(&term, B9600);
    cfsetospeed(&term, B9600);
#endif

    tcsetattr(fd, TCSANOW, &term);

    gettimeofday(&time, NULL);
    st = time.tv_sec;

    /* 
     * sending break to disconnect and waiting carrier down
     */

    do
    {
	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "(hangup) %s setting DTR.\n", tname);
	setdtr(fd, 1);

	if (!mp || mp->hang_method == HUP_BREAK)
	{
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "(hangup) %s sending break.\n", tname);

	    /* send break signal */
	    if (tcsendbreak(fd, BREAKTIME) == -1)
	    {
		writemsg(SYSTM, "(hangup) sendbreak error: %s\n",
							strerror(errno));
		break;
	    }
	}
	else
	    lsleep(1);

	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "(hangup) %s clearing DTR.\n", tname);
	setdtr(fd, 0);

	if (mp && mp->hang_method != HUP_BREAK)
	    lsleep(5);

	if (trace & TRACE_CTRL)
	    writemsg(TRACE, "(hangup) %s setting DTR.\n", tname);
	setdtr(fd, 1);

	/* flush all output */
	if (tcflush(fd, TCIOFLUSH) == -1)
	{
	    writemsg(SYSTM, "(hangup) flush error: %s\n", strerror(errno));
	    break;
	}

	if (mp->hang)
	    break;

	/* check carrier*/
	if (!getcar(fd))
	    break;

	/* waiting carrier down */
	if (trace & TRACE_MODEM)
	    dumpmctrl(fd);
	else if (trace & TRACE_CTRL)
	    writemsg(TRACE, "(hangup) waiting carrier down.\n");

	lsleep(5);
	gettimeofday(&time, NULL);
    }
    while(time.tv_sec - st < 2);

#ifdef RPORT_RESET
    /*
     *  recover baud rate
     */
    cfsetispeed(&term, ispeed);
    cfsetospeed(&term, ospeed);

#endif
    tcsetattr(fd, TCSANOW, &term);

    /*
     * waiting escape mode prompt
     */
    if (mp->esc)
    {
	/* sending hsync command because line may not connected */
	if (write(fd, mp->hsync, strlen(mp->hsync)) < 0
	 || write(fd, mcr, sizeof(mcr) - 1) < 0)
	    return -1;

	do
	{
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "(hangup) waiting escape mode prompt.\n");

	    if (!(resp = mgets(fd, MMWAIT)))
	    {
		if (trace & TRACE_CTRL)
		    writemsg(TRACE, "(hangup) escape timeout.\n");
		break;
	    }

	    /* response mp->esc */
	} while(strncmp(resp, mp->esc, strlen(mp->esc)));
    }

    /*
     * sending hangup string to disconnect
     */
    if (mp->hang)
    {
	/* 
	 * sending hangup string to disconnect and waiting response.
	 */

	char **rcp;

	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "(hangup) sending '%s'.\n", mp->hang);

	/* sending hangup string */

	if (write(fd, mp->hang, strlen(mp->hang)) < 0
	 || write(fd, mcr, sizeof(mcr) - 1) < 0)
	    return -1;

	tcdrain(fd);

	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "(hangup) waiting hangup response.\n");
	/*
	 * reset terminal attribute
	 */
	tcsetattr(fd, TCSANOW, &term);

	do
	{
	    if (!(resp = mgets(fd, MMWAIT)))
	    {
		if (trace & TRACE_MODEM)
		    writemsg(TRACE, "(hangup) hangup timeout.\n");
		break;
	    }

	    /* response OK */
	    for(rcp = r_ok; *rcp; rcp++)
		if (!strncmp(resp, *rcp, strlen(*rcp)))
		{
		    if (trace & TRACE_MODEM)
			writemsg(TRACE, "(hangup) hanged up.\n");
		    break;
		}

	    if (*rcp)
	    	break;

	    /* response NO CARRIER */
	    for(rcp = r_busy; *rcp; rcp++)
		if (!strncmp(resp, *rcp, strlen(*rcp)))
		{
		    if (trace & TRACE_MODEM)
			writemsg(TRACE, "(hangup) disconnected\n");
		    break;
		}

	} while(!*rcp);
    }


    /*
     * synchronize modem and host
     */
    if (mp && mp->hsync)
    {
	char **rcp;

	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "(hangup) hangup synchronization.\n");

	lsleep(2);

	if (trace & TRACE_MODEM)
	    writemsg(TRACE, "(hangup) sending '%s'.\n", mp->hsync);

	if (write(fd, mp->hsync, strlen(mp->hsync)) < 0
	 || write(fd, mcr, sizeof(mcr) - 1) < 0)

	    /* hsync characters may cause hangup(I/O error). */
	    return hangup(fd, tty, mp);

	tcdrain(fd);

	do
	{
	    char *resp;

	    if (!(resp = mgets(fd, MMWAIT)))
	    {
		writemsg(TRACE, "(hangup) hangup sync. timeout.\n");
		break;
	    }

	    /* response OK */
	    for(rcp = r_ok; *rcp; rcp++)
		if (!strncmp(resp, *rcp, strlen(*rcp)))
		    break;
	} while(!*rcp);
    }

    if (tcflush(fd, TCIOFLUSH) == -1)
	writemsg(SYSTM, "(hangup) flush error: %s\n", strerror(errno));

    /* when modem doesn't close connection by break signal */
    if (getcar(fd))
    	return -1;

    /*
     * signal when modem break
     */
    term.c_iflag &= ~IGNBRK;
    term.c_iflag |=  BRKINT;
    term.c_cflag = cflag;
    tcsetattr(fd, TCSANOW, &term);

    if (trace & TRACE_MODEM)
    {
	tcgetattr(fd, &term);
	ospeed = cfgetospeed(&term);
	ispeed = cfgetispeed(&term);
	writemsg(TRACE, "(hangup) ispeed = 0x%x, ospeed = 0x%x.\n",
							ispeed, ospeed);
    }

    if (trace & TRACE_CTRL)
	writemsg(TRACE, "(hangup) finished.\n");

    return 0;
}

void
trace_modem(fi, fo, tty, modem, type, env)
    int fi, fo;
    char *tty;
    struct mdm *modem;
    int type;
    struct vars *env;
{
    int i;
    char cbuf[MAXFLINE];

    for(i = 0; i < MAXTRACE; i++)
	if (modem->trace[i].type & type)
	{
	    expandenv(cbuf, MAXFLINE, modem->trace[i].name, env);
	    mset(fi, fo, modem->trace[i].modem, tty, cbuf);
	}
}


void
dumpmctrl(fd)
    int fd;
{
    int mctrl;

    ioctl(fd, TIOCMGET, &mctrl);
    writemsg(TRACE, "MODEM LINE: %s %s %s %s %s %s %s %s %s\n",
				mctrl & TIOCM_LE ?"LE" :"le",
				mctrl & TIOCM_DTR?"DTR":"dtr",
				mctrl & TIOCM_RTS?"RTS":"rts",
				mctrl & TIOCM_ST ?"ST" :"st",
				mctrl & TIOCM_SR ?"SR" :"sr",
				mctrl & TIOCM_CTS?"CTS":"cts",
				mctrl & TIOCM_CAR?"CAR":"car",
				mctrl & TIOCM_RNG?"RNG":"rng",
				mctrl & TIOCM_DSR?"DSR":"dsr");
}

int
getcar(fd)
    int fd;
{
    int mctrl;

    if (ioctl(fd, TIOCMGET, &mctrl) == -1)
    {
    	writemsg(SYSTM, "Error reading modem control line: %s\n",
	    						strerror(errno));
    	mctrl = 0;
    }
    return mctrl & TIOCM_CAR;
}

int
setdtr(fd, on)
    int fd;
    int on;
{
    int mctrl;

    if (ioctl(fd, TIOCMGET, &mctrl) == -1)
    {
    	writemsg(SYSTM, "Error reading modem control line: %s\n",
	    						strerror(errno));
    	return -1;
    }

    if (on)
	mctrl |= TIOCM_DTR;
    else
	mctrl &= ~TIOCM_DTR;

    if (ioctl(fd, TIOCMSET, &mctrl) == -1)
    {
    	writemsg(SYSTM, "Error setting modem control line: %s\n",
	    						strerror(errno));
    }

    return 0;
}

int
waitfd(fd, top)
    int fd;
    struct timeval *top;
{
    int ret;
#ifdef FD_SET
    fd_set fds;
#else
    struct pollfd fds[1];
#endif

    do
    {
    	trace_to = 0;

#ifdef FD_SET

	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	ret = select(fd + 1, &fds, NULL, NULL, top);
#else

	fds[0].fd = fd;
	fds[0].events = POLLIN;
# ifdef INFTM
#  define INFTM -1
# endif
	ret = poll(fds, 1, top ? top->tv_sec * 1024 + top->tv_usec / 1024:
				 INFTM);
#endif

    } while(ret == -1 && trace_to);

    return ret;
}

/*
 *	functions defined for dcba
 */

char*
waitcolon(fd, sec)
    int fd;
    int sec;
{
    char c;
    int  r;
    struct timeval to;
    struct termios term, torg;
    static char cbuf[MAXFLINE];
    char *cp;

    tcgetattr(fd, &torg);
    memcpy(&term, &torg, sizeof(struct termios));

    term.c_lflag    &= ~ICANON;

    term.c_cc[VTIME] = 0;
    term.c_cc[VMIN]  = 1;

    tcsetattr(fd, TCSANOW, &term);

    cp = cbuf;

    c = NUL;

    do
    {
	to.tv_sec = sec;
	to.tv_usec = 0;
    	r = waitfd(fd, &to);
	if (r == 0)
	{
	    tcsetattr(fd, TCSANOW, &torg);
	    writemsg(SYSTM, "timeout waiting ':'\n");
	    return NULL;
	}
	if (r < 0)
	{
	    tcsetattr(fd, TCSANOW, &torg);
	    writemsg(SYSTM, "select/poll error: %s\n", strerror(errno));
	    return NULL;
	}

	r = read(fd, &c, 1);
	if (r < 0)
	{
	    tcsetattr(fd, TCSANOW, &torg);
	    writemsg(SYSTM, "read error: %s\n", strerror(errno));
	    return NULL;
	}

	if (c == '\n' || c == '\r')
	    break;

	if (c >= ' ' && !(c & 0x80))
	    *cp++ = c;

	if (cp - cbuf > MAXFLINE - 1)
	    break;

	to.tv_sec = 0;
	to.tv_usec = 0;
	if ((c == ':' || c == ' ') && waitfd(fd, &to) == 0)
	    break;

    }
    while(1);

    *cp = NUL;
    tcsetattr(fd, TCSANOW, &torg);
    return cbuf;
}

int
waitcarrier(fd, on)
    int fd;
    int on;
{
    int car;
    int r;
    struct timeval to;
    struct termios term, torg;

    tcgetattr(fd, &torg);
    memcpy(&term, &torg, sizeof(struct termios));

    term.c_lflag    &= ~ICANON;

    term.c_cc[VTIME] = 0;
    term.c_cc[VMIN]  = 1;

    tcsetattr(fd, TCSANOW, &term);

    do
    {
	car = getcar(fd);

	if (on ? car: !car)
	    break;

	to.tv_sec = 5;
	to.tv_usec = 0;
    	r = waitfd(fd, &to);
	if (r == 0)
	    return 1;
	if (r < 0)
	{
	    writemsg(SYSTM, "select/poll error: %s\n", strerror(errno));
	    tcsetattr(fd, TCSANOW, &torg);
	    return -1;
	}
    }
    while(r);

    tcsetattr(fd, TCSANOW, &torg);
    return 1;
}

void
response(fd, cp)
    int fd;
    char *cp;
{
    write(fd, cp, strlen(cp));
    write(fd, "\n", 1);
}

