/* vi:set ai sm

	utmp, wtmp management routine
				by ann@mrit.co.jp

*/


#include "cba.h"


/* macros for statistics */
#ifndef MAXTTY
#define MAXTTY	128	/* maximum tty  */
#endif

#ifdef MAXPHONES
# define MAXUSR	(MAXPHONES + 1)	/* maximum user */
#else
# define MAXUSR	512	/* maximum user */
#endif

#define UTBUFSIZE 128	/* how many entries copied at once */

extern int trace;
extern writemsg();

static void putut();

#if defined(UTMPX_FILE)
# define ut_time ut_tv.tv_sec
#endif

int
addutmp(file, tty, name, from)
    char *file;
    char *tty, *name, *from;
{
#ifdef UTMPX_FILE
    struct utmpx ut;
#else
    struct utmp ut;
#endif
    int fut;
    int pos, epos;
    char *cp;

    if (cp = strchr(tty + 1, '/'))
	tty = cp + 1;

    if (trace & TRACE_ULOG)
	writemsg(TRACE, "add user %s@%s (%s)\n", name, tty, file);

    fut = open(file, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    if (fut == EOF)
	return -1;

    LOCKFILE(fut);

    epos = -1;
    do
    {
	pos = lseek(fut, 0L, SEEK_CUR); 	/* get current posission */
        if (read(fut, &ut, sizeof(ut)) <= 0)
	{
	    /* could not found */
	    pos = epos;
	    break;
	}

        if (! *ut.ut_line)
	    /* empty slot found */
	    epos = pos;

    } while(strcmp(ut.ut_line, tty));

    if (pos == -1)
	lseek(fut, 0L, SEEK_END);	/* append new slot */
    else
	lseek(fut, pos, SEEK_SET);

    putut(fut, tty, name, from);

    UNLKFILE(fut);

    close(fut);

    return pos;
}

int
addwtmp(file, tty, name, from)
    char *file;
    char *tty, *name, *from;
{
    int fut;
    char *cp;

    if (cp = strchr(tty + 1, '/'))
	tty = cp + 1;

    fut = open(file, O_RDWR | O_CREAT | O_APPEND, 
					S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    if (fut == EOF)
    {
	writemsg(SYSTM, "wtmp file %s open failed.\n", file);
	return -1;
    }

    LOCKFILE(fut);
    putut(fut, tty, name, from);
    UNLKFILE(fut);

    close(fut);

    if (trace & TRACE_ULOG)
	if (name)
	    writemsg(TRACE, "%s@%s logged in(%s)\n", name, tty, file);
	else
	    writemsg(TRACE, "%s released (%s)\n", tty, file);

    return 0;
}

int
rmutmp(file, tty)
    char *file;
    char *tty;
{
#ifdef UTMPX_FILE
    struct utmpx ut;
#else
    struct utmp ut;
#endif
    int fut;
    int pos;
    char *cp;

    if (cp = strchr(tty + 1, '/'))
	tty = cp + 1;

    fut = open(file, O_RDWR);

    if (fut == EOF)
    {
	writemsg(SYSTM, "utmp file %s open failed.\n", file);
	return -1;
    }

    LOCKFILE(fut);

    do
    {
	int n;

	pos = lseek(fut, 0L, SEEK_CUR); 	/* get current posission */
        n = read(fut, &ut, sizeof(ut));
        if (n <= 0)
	{
	    if (trace & TRACE_ULOG)
		writemsg(TRACE, "%s not found in %s.\n", tty, file);
	    UNLKFILE(fut);
	    close(fut);
	    return -1;
	}
    } while(strcmp(ut.ut_line, tty));

    lseek(fut, pos, SEEK_SET);

    putut(fut, NULL, NULL, NULL);
    UNLKFILE(fut);
    close(fut);
    return 0;
}

/*
void
pututmp(file, pos, tty, name, from)
    char *file;
    int pos;
    char *tty, *name, *from;
{
    int    fut;

    fut = open(file, O_RDWR);
    if (fut == EOF)
    {
	writemsg(SYSTM, "utmp file %s open failed.\n", file);
	return;
    }

    LOCKFILE(fut);
    lseek(fut, pos, SEEK_SET);

    putut(fut, tty, name, from);
    UNLKFILE(fut);
    close(fut);
}
*/

static void
putut(fut, tty, name, from)
    int fut;
    char *tty, *name, *from;
{
#ifdef UTMPX_FILE
    struct utmpx ut;
    char *dp;
    extern char *ver;
#else
    struct utmp ut;
#endif
    char *cp;
    int i;
    struct timeval now;

    gettimeofday(&now, NULL);

    memset(ut.ut_line, NUL, sizeof(ut.ut_line));
    if (tty)
    {
	if (*tty == '/')
	    cp = (cp = strchr(tty + 1, '/'))? cp + 1 : tty;
	else
            cp = tty;

#ifdef UTMPX_FILE
    	if (*tty == '~')	/* for reboot */
	    cp = BOOT_MSG;
#endif
	if (sizeof(ut.ut_line) < strlen(cp))
	    strncpy(ut.ut_line, cp, sizeof(ut.ut_line));
	else
	    strcpy(ut.ut_line, cp);
    }

    memset(ut.ut_name, NUL, sizeof(ut.ut_name));
    if (name)
    {
	if (sizeof(ut.ut_name) < strlen(name))
	    strncpy(ut.ut_name, name, sizeof(ut.ut_name));
	else
	    strcpy(ut.ut_name, name);
    }

    memset(ut.ut_host, NUL, sizeof(ut.ut_host));
    if (from)
    {
	if (sizeof(ut.ut_host) < strlen(from))
	    strncpy(ut.ut_host, from, sizeof(ut.ut_host));
	else
	    strcpy(ut.ut_host, from);
    }

#if defined(UTMPX_FILE)
    ut.ut_syslen = strlen(ut.ut_host);
    if (ut.ut_syslen > sizeof(ut.ut_host))
    	ut.ut_syslen = sizeof(ut.ut_host);
    ut.ut_tv = now;
    ut.ut_exit.e_termination =
    ut.ut_exit.e_exit = 0;
    ut.ut_pid = getpid();
    memset(ut.ut_id, '_', sizeof(ut.ut_id));
    dp = ut.ut_id;
    for(cp = "cba"; *cp && dp < ut.ut_id + sizeof(ut.ut_id); dp++, cp++)
	*dp = *cp;
    for(cp = ver; *cp && dp < ut.ut_id + sizeof(ut.ut_id); dp++, cp++)
	*dp = *cp;
#else
    ut.ut_time = now.tv_sec;
#endif

#ifdef BOOT_TIME
    if (!tty)
# if defined(UT_UNKNOWN)
    	ut.ut_type = UT_UNKNOWN;
# elif defined(EMPTY)
    	ut.ut_type = EMPTY;
# else
    	ut.ut_type = 0;
# endif
    else if (*tty == '~')	/* for reboot */
	ut.ut_type = BOOT_TIME;
    else
	ut.ut_type = USER_PROCESS;
#endif

    write(fut, &ut, sizeof(ut));
}

void
dmputmp(file)
    char *file;
{
#ifdef UTMPX_FILE
    struct utmpx ut;
#else
    struct utmp ut;
#endif
    int    fut;
    int    i;
    char   *cp;

    fut = open(file, O_RDONLY);
    if (fut == EOF)
    {
	writemsg(SYSTM, "utmp file %s open failed.\n", file);
	return;
    }

    LOCKFILE(fut);

    while(read(fut, &ut, sizeof(ut))>0)
    {
	if (!*ut.ut_line)
	    continue;

	i = sizeof(ut.ut_line);
	if (i > 12)
	    i = 12;
	for(cp = ut.ut_line; i; cp++, i--)
	    fputc(isprint(*cp)? *cp: ' ' , stdout);
	fputc(' ', stdout);

	i = sizeof(ut.ut_name);
	if (i > 8)
	    i = 8;
	for(cp = ut.ut_name; i; cp++, i--)
	    fputc(isprint(*cp)? *cp: ' ' , stdout);
	fputc(' ', stdout);

	i = sizeof(ut.ut_host);
	if (i > 16)
	    i = 16;
	for(cp = ut.ut_host; i; cp++, i--)
	    fputc(isprint(*cp) || *cp & 0x80 ? *cp: ' ' , stdout);

#ifdef UTMPX_FILE
	fputc(' ', stdout);

	for(cp = ut.ut_id, i = sizeof(ut.ut_id); i; cp++, i--)
	    fputc(isprint(*cp)? *cp: ' ' , stdout);
#endif
	fprintf(stdout, " %s", ut.ut_time ? ctime(&ut.ut_time): "-\n");
    }
    UNLKFILE(fut);
    close(fut);
}


static time_t
str2time(tstr)
    char *tstr;
{
    struct timeval tv;
    struct tm tmret, *tmp;
    time_t now;
    char *cp;
    int  i;

    gettimeofday(&tv, NULL);
    now = (time_t)tv.tv_sec;
    tmp = localtime(&now);

    tmret.tm_isdst  = 0;
    tmret.tm_sec  = 0;
    tmret.tm_min  = 0;
    tmret.tm_hour = 0;
    tmret.tm_mday = tmp-> tm_mday;
    tmret.tm_mon  = tmp-> tm_mon;
    tmret.tm_year = tmp-> tm_year;

    /* 
     *	yy/mm/dd
     *	yy/mm/dd,hh:mm
     *	   mm/dd
     *	   mm/dd,hh:mm
     *	      dd
     *	      dd,hh:mm
     *	         hh:mm
     */

    for(cp = tstr, i = 0; *cp; cp++)
    	if (*cp == '/')
	    i++;

    switch(i)
    {
    case 2:	/* yy/mm/dd */
    	cp = tstr;
	for(;*tstr != '/'; tstr++);
	*tstr ++ = NUL;
	tmret.tm_year = atoi(cp);

    case 1:	/*    mm/dd */
    	cp = tstr;
	for(;*tstr != '/'; tstr++);
	*tstr ++ = NUL;
	tmret.tm_mon = atoi(cp) - 1;

    case 0:	/*       dd , dd,hh:mm or hh:mm  */
	for(cp = tstr, i = 0; *cp; cp++)
	    if (*cp == ',')
		i++;
	    else if (*cp == ':')
	    	break;
	if (*cp == ':' && i == 0)
	    break;

    	cp = tstr;
	for(;*tstr >= '0' && *tstr <= '9'; tstr++);
	if (*tstr)
	    *tstr ++ = NUL;
	tmret.tm_mday = atoi(cp);
    default:
    	break;
    }

    for(cp = tstr, i = 0; *cp; cp++)
    	if (*cp == ':')
	    i++;

    if (i)	/* hh:mm */
    {
    	cp = tstr;
	for(;*tstr != ':'; tstr++);
	*tstr ++ = NUL;
	tmret.tm_hour = atoi(cp);

    	cp = tstr;
	for(;*tstr; tstr++);
	*tstr ++ = NUL;
	tmret.tm_min = atoi(cp);
    }

    return mktime(&tmret);
}


void
dmpwtmp(file, start, end, user)
    char *file;
    char *start, *end;
    char *user;
{
#ifdef UTMPX_FILE
    struct utmpx ut;
#else
    struct utmp ut;
#endif
    struct tm *tmp;
    struct timeval tv;
    time_t time_s, time_e, now;
    int    fut;
    int    ntty;
    int    nlogin;

    int  i;
    char ttyname[MAXTTY][sizeof(ut.ut_line) + 1];
    char usrname[MAXUSR][sizeof(ut.ut_name) + 1];
    char phoname[MAXUSR][sizeof(ut.ut_host) + 1];
    time_t  prev[MAXTTY], sum_prev;
    int  avail[MAXTTY];
    int  tty_a[MAXTTY], tty_u[MAXTTY];
    int  sum[MAXTTY];
    int  uid[MAXTTY];
    int  tid[MAXTTY];	/* uid for total user */
    int  utime[MAXUSR];
    int  ttime[MAXUSR];

    time_s = str2time(start);
    time_e = str2time(end);

    gettimeofday(&tv, NULL);
    now = (time_t)tv.tv_sec;

    if (time_e > now)
    	time_e = now;

    tmp = localtime(&time_s);
    printf("%d/%02d/%02d %2d:%02d - ",
    		1900 + tmp->tm_year, tmp->tm_mon + 1, tmp->tm_mday,
		tmp->tm_hour, tmp->tm_min);

    tmp = localtime(&time_e);
    printf("%d/%02d/%02d %2d:%02d",
    		1900 + tmp->tm_year, tmp->tm_mon + 1, tmp->tm_mday,
		tmp->tm_hour, tmp->tm_min);

    if (user)
	printf("\tfor user %s", user);

    printf("\n");

    for(i = 0; i < MAXTTY; i++)
    {
        ttyname[i][0] = NUL;
	prev[i] = sum_prev = time_s;
	avail[i] = 1;
	uid[i] =
	tid[i] =
	sum[i] =
	tty_a[i] =
	tty_u[i] = 0;
    }
    ntty = 0;
    nlogin = 0;

    for(i = 0; i < MAXUSR; i++)
    {
        ttime[i] =
        utime[i] = 0;
        usrname[i][0] = NUL;
        phoname[i][0] = NUL;
    }

    strcpy(&usrname[0][0], "unknown");
    strcpy(&phoname[0][0], "?");

    if (file)
    	fut = open(file, O_RDONLY);
    else
	fut = 0;

    if (fut < 0)
    {
    	fprintf(stderr, "file %s cannot open\n", file);
	return;
    }

    LOCKFILE(fut);

    printf("\n");

    while(read(fut, &ut, sizeof(ut)) > 0)
    {
	if (ut.ut_time < time_s)
	{
	    /*
	     *	do not compute time.
	     */
#ifdef BOOT_TIME
	    if (ut.ut_type == BOOT_TIME
	     || ut.ut_line[0] == '~')	/* shutdown */
#else
	    if (ut.ut_line[0] == '~')	/* shutdown */
#endif
	    {
		for(i = 0; i < MAXTTY; i++)
		    avail[i] = 1;
		ntty = 0;
	    }
	    else if (ut.ut_name[0])	/* login */
	    {
		for(i = 0; i < MAXTTY; i++)
		    if (ttyname[i][0] == NUL)
		    {
			strncpy(&ttyname[i][0], ut.ut_line, sizeof(ut.ut_line));
			ttyname[i][sizeof(ut.ut_line) + 1] = NUL;
			break;
		    }
		    else if (!strncmp(&ttyname[i][0], ut.ut_line,
		    					sizeof(ut.ut_line)))
			break;
		
		if (i < MAXTTY)
		{
		    int j;

		    ntty ++;
		    tid[i] =
	    	    avail[i] = 0;
		    for(j = 0; j < MAXUSR; j++)
		    	if (usrname[j][0] == NUL)
			{
			    strncpy(&usrname[j][0], ut.ut_name,
			    				sizeof(ut.ut_name));
			    usrname[j][sizeof(ut.ut_name) + 1] = NUL;

			    strncpy(&phoname[j][0], ut.ut_host,
			    				sizeof(ut.ut_host));
			    phoname[j][sizeof(ut.ut_host) + 1] = NUL;
			    uid[i] = j;
			    if (tid[i] == 0)
			    	tid[i] = j;
			    break;
			}
			else if (!strncmp(&usrname[j][0], ut.ut_name,
		    					sizeof(ut.ut_name)))
			{
			    if (tid[i] == 0)
			    	tid[i] = j;
			    if (!strncmp(&phoname[j][0], ut.ut_host,
		    					sizeof(ut.ut_host)))
			    {
				uid[i] = j;
				break;
			    }
			}
		}
	    }
	    else			/* logout */
	    {
		for(i = 0; i < MAXTTY; i++)
		    if (!strncmp(&ttyname[i][0], ut.ut_line,
		    					sizeof(ut.ut_line)))
			break;
		if (i < MAXTTY)
		{
		    avail[i] = 1;
		    ntty --;
		    if (ntty < 0)
		    	ntty = 0;
		}
	    }
	    continue;
	}

	if (ut.ut_time > time_e)
	    break;

	/*
	 *	compute time.
	 */

#ifdef BOOT_TIME
	if (ut.ut_type == BOOT_TIME
	 || ut.ut_line[0] == '~')	/* shutdown */
#else
	if (ut.ut_line[0] == '~')	/* shutdown */
#endif
	{
	    for(i = 0; i < MAXTTY; i++)
	    {
		if (ttyname[i][0] == NUL)
		    break;

		if (avail[i])
		    tty_a[i] += (ut.ut_time - prev[i]);
		else
		{
		    tty_u[i] += (ut.ut_time - prev[i]);
		    utime[uid[i]] += (ut.ut_time - prev[i]);
		    ttime[tid[i]] += (ut.ut_time - prev[i]);
		}

		if (user && !strcmp(usrname[uid[i]], user))
		{
		    struct tm *tmp;
		    time_t    t;

		    printf("%8s@%-16s %-6s ",	&usrname[uid[i]][0],
						&phoname[uid[i]][0],
						ttyname[i]);

		    if (prev[i] == time_s)
			printf("               - ");
		    else
		    {
			tmp = localtime(&prev[i]);
			printf("%2d/%02d %2d:%02d:%02d - ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
		    }

		    tmp = localtime(&ut.ut_time);
		    printf("%2d/%02d %2d:%02d:%02d ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);

		    t = ut.ut_time - prev[i];
		    printf("%4d:%02d:%02d(shutdown)\n",
					t/3600, (t%3600)/60, t%60);

		}

		avail[i] = 1;
		prev[i] = ut.ut_time;
	    }
	    if (ntty >= 0)
		sum[ntty] += (ut.ut_time - sum_prev);
	    sum_prev = ut.ut_time;
	    ntty = 0;

	    if (trace & TRACE_TTY)
		printf("~");
	}

	else if (ut.ut_name[0])		/* login */
	{
	    int j;

	    nlogin ++;

	    for(i = 0; i < MAXTTY; i++)
		if (ttyname[i][0] == NUL)
		{
		    strncpy(&ttyname[i][0], ut.ut_line, sizeof(ut.ut_line));
		    ttyname[i][sizeof(ut.ut_line) + 1] = NUL;
		    break;
		}
		else if (!strncmp(&ttyname[i][0], ut.ut_line, sizeof(ut.ut_line)))
		    break;
	    if (i >= MAXTTY)
	    {
	    	fprintf(stderr, "MAXTTY too small. recompile this program.\n");
		continue;
	    }


	    if (ntty >= 0)
		sum[ntty] += (ut.ut_time - sum_prev);

	    if (!avail[i])	/* duplicated login */
	    {
		tty_u[i] += (ut.ut_time - prev[i]);
	    	utime[0] += (ut.ut_time - prev[i]);

		if (trace & TRACE_TTY)
		    printf("+");
		else
		{
		    struct tm *tmp;
		    int t;

		    tmp = localtime(&ut.ut_time);
		    printf("!%2d/%02d %2d:%02d:%02d ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);

		    printf("%s overwrites %s@%s ", ut.ut_name,
						&usrname[i][0],
						&phoname[i][0]);

		    tmp = localtime(&prev[i]);
		    printf("%2d/%02d %2d:%02d:%02d ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);

		    t = ut.ut_time - prev[i];
		    printf("%2d:%02d:%02d ",
					t/3600, (t%3600)/60, t%60);

		    printf("%s\n", &ttyname[i][0]);
		}
	    }
	    else
	    {
		tty_a[i] += (ut.ut_time - prev[i]);
		if (trace & TRACE_TTY)
		    printf(" ");
	    	ntty += 1;
	    }
	    avail[i] = 0;
	    sum_prev =
	    prev[i] = ut.ut_time;


	    tid[i] = 0;
	    for(j = 0; j < MAXUSR; j++)
	    	if (usrname[j][0] == NUL)
		{
		    strncpy(&usrname[j][0], ut.ut_name, sizeof(ut.ut_name));
		    usrname[j][sizeof(ut.ut_name) + 1] = NUL;

		    strncpy(&phoname[j][0], ut.ut_host, sizeof(ut.ut_host));
		    phoname[j][sizeof(ut.ut_host) + 1] = NUL;
		    uid[i] = j;
		    if (tid[i] == 0)
		    	tid[i] = j;
		    break;
		}
		else if (!strncmp(&usrname[j][0],ut.ut_name,sizeof(ut.ut_name)))
		{
		    if (tid[i] == 0)
		    	tid[i] = j;
		    if (!strncmp(&phoname[j][0], ut.ut_host,sizeof(ut.ut_host)))
		    {
			uid[i] = j;
			break;
		    }
		}
	}

	else				/* logout */
	{
	    for(i = 0; i < MAXTTY; i++)
		if (!strncmp(&ttyname[i][0], ut.ut_line, sizeof(ut.ut_line)))
		    break;
	    if (i >= MAXTTY)
	    	continue;

	    if (avail[i])	/* duplicated logout */
	    {
		if (trace & TRACE_TTY)
		    printf("-");
		else
		{
		    struct tm *tmp;

		    tmp = localtime(&ut.ut_time);
		    printf("!%2d/%02d %2d:%02d:%02d ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);

		    printf("empty logout ");

		    tmp = localtime(&prev[i]);
		    printf("(last %2d/%02d %2d:%02d:%02d) ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);

		    printf("%s\n", &ttyname[i][0]);
		}
	    }
	    else
	    {
	    	int t;

		if (trace & TRACE_TTY)
		    printf(".");
		avail[i] = 1;
		sum[ntty] += (ut.ut_time - sum_prev);

		t = ut.ut_time - prev[i];
		tty_u[i]      += t;
	    	utime[uid[i]] += t;
		ttime[tid[i]] += t;

		if ((user && !strcmp(usrname[uid[i]], user))
		   || trace & TRACE_CTRL)
		{
		    struct tm *tmp;

		    printf("%8s@%-16s %-6s ",	&usrname[uid[i]][0],
						&phoname[uid[i]][0],
						ttyname[i]);

		    if (prev[i] == time_s)
			printf("               - ");
		    else
		    {
			tmp = localtime(&prev[i]);
			printf("%2d/%02d %2d:%02d:%02d - ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
		    }
		    tmp = localtime(&ut.ut_time);
		    printf("%2d/%02d %2d:%02d:%02d ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);

		    printf("%4d:%02d:%02d\n",
					t/3600, (t%3600)/60, t%60);

		}
	    }

	    sum_prev =
	    prev[i] = ut.ut_time;

	    ntty -= 1;
	    if (ntty < 0)
		ntty = 0;
	}

	if (trace & TRACE_TTY)
	{
	    tmp = localtime(&ut.ut_time);
	    printf(" %2d/%2d %2d:%02d(%2d) ",	tmp->tm_mon + 1, tmp->tm_mday,
						tmp->tm_hour, tmp->tm_min,
						ntty);

	    for(i = 0; i < MAXTTY; i++)
		if (ttyname[i][0])
		    printf("%c", avail[i]? ' ': '*');

	    printf("\n");
	}
    }

    UNLKFILE(fut);
    close(fut);

    for(i = 0; i < MAXTTY; i++)
    	if (!ttyname[i][0])
	    break;
	else if (avail[i])
	    tty_a[i] += (time_e - prev[i]);
	else
	{
	    time_t t;

	    t = time_e - prev[i];

	    tty_u[i] += t;
	    ttime[tid[i]] += t;
	    utime[uid[i]] += t;

	    if ((user && !strcmp(usrname[uid[i]], user))
		|| trace & TRACE_CTRL)
	    {
		struct tm *tmp;

		printf("%8s@%-16s %-6s ",	&usrname[uid[i]][0],
						&phoname[uid[i]][0],
						ttyname[i]);

		if (prev[i] == time_s)
		    printf("               - ");
		else
		{
		    tmp = localtime(&prev[i]);
		    printf("%2d/%02d %2d:%02d:%02d - ",
				tmp->tm_mon + 1, tmp->tm_mday,
				tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
		}
		printf("still logged in");

		printf("%4d:%02d:%02d\n", t/3600, (t%3600)/60, t%60);

	    }
	}

    if (ntty >= 0)
	sum[ntty] += (time_e - sum_prev);

    printf("\n");

    {
    	int tuse;
        int utotal;
        int ptotal;
	int maxtty;

	if (!user)
	{
	    printf("---- tty statistics ----------------------\n");

	    tuse = 0;
	    for(i = 0; i < MAXTTY; i++)
	    {
		if (!ttyname[i][0])
		    break;

		if (tty_u[i] == 0)
		    continue;

		printf(" %s\t%4d:%02d:%02d used. \t %4d:%02d:%02d free.\n",
			&ttyname[i][0],
			tty_u[i]/3600, (tty_u[i]%3600)/60, tty_u[i]%60,
			tty_a[i]/3600, (tty_a[i]%3600)/60, tty_a[i]%60);
		tuse += tty_u[i];
	    }
	}

	if (!user)
	{
	    printf("\n");
	    printf("---- tty availability ------------------\n");
	    for(i = 0; i < MAXTTY; i++)
	    {
		if (sum[i] == 0)
		    continue;

		printf(" %2d ttys used: %4d:%02d:%02d\n",
				i, sum[i]/3600, (sum[i]%3600)/60, sum[i]%60);
		maxtty = i;
	    }
	}

	printf("\n");
	if (user)
	    printf("---- total -------------------------------\n");
	else
	    printf("---- user statistics ---------------------\n");

	utotal =
	ptotal = 0;

	for(i = 0; i < MAXUSR; i++)
	{
	    if (!usrname[i][0])
		break;

	    if (user && strcmp(usrname[i], user))
		continue;

	    if (utime[i] == 0)
	    {
	        int j;

	    	if (ttime[i])
		    for(j = i + 1; j < MAXUSR; j++)
			if (!strcmp(&usrname[i][0], &usrname[j][0]))
			{
			    ttime[j] = ttime[i];
			    break;
			}
		continue;
	    }

	    if (i != 0)
	    {
		ptotal ++;
		if (ttime[i] != 0)
		    utotal ++;
	    }

	    printf(" %8s@%-16s %4d:%02d:%02d",
			&usrname[i][0],
			&phoname[i][0],
			utime[i]/3600, (utime[i]%3600)/60, utime[i]%60);

	    if (ttime[i] && (!user || ttime[i] != utime[i]))
		printf(" (total %4d:%02d:%02d)",
			ttime[i]/3600, (ttime[i]%3600)/60, ttime[i]%60);

	    printf("\n");
	}

	if (!user)
	{
	    printf("\n");
	    printf("total: %4d:%02d:%02d  %d people %d place %d tty %d callback\n",
					tuse/3600, (tuse%3600)/60, tuse%60,
					utotal, ptotal, maxtty, nlogin);
	}
    }
}


void
cutwtmp(file, newfile, limit)
    char *file, *newfile;
    char *limit;
{
#ifdef UTMPX_FILE
    struct utmpx ut, utbuf[UTBUFSIZE];
#else
    struct utmp ut, utbuf[UTBUFSIZE];
#endif
    struct tm *tmp;
    time_t time;
    int    fut;
    int    fnew;
    int	   pos;
    int    cpos;
    int    ntty;

    int  i;

    char ttyname[MAXTTY][sizeof(ut.ut_line) + 1];
    int  avail[MAXTTY];

    time = str2time(limit);

    for(i = 0; i < MAXTTY; i++)
    {
        ttyname[i][0] = NUL;
	avail[i] = 1;
    }
    fut = open(file, O_RDONLY);
    if (fut < 0)
    {
    	fprintf(stderr, "file %s cannot open\n", file);
	return;
    }

    pos = lseek(fut, 0L, SEEK_CUR); 	/* get current posission */
    ntty = 0;

    while(read(fut, &ut, sizeof(ut)) > 0)
    {
	if (ut.ut_time >= time)
	    break;

	cpos = lseek(fut, 0L, SEEK_CUR); 	/* get current posission */

	/*
	 *	do not compute time.
	 */
#ifdef BOOT_TIME
	if (ut.ut_type == BOOT_TIME
	 || ut.ut_line[0] == '~')	/* shutdown */
#else
	if (ut.ut_line[0] == '~')	/* shutdown */
#endif
	{
	    for(i = 0; i < MAXTTY; i++)
		avail[i] = 1;
	    pos = cpos;
	    ntty = 0;
	    if (trace)
		printf("~");
	}
	else if (ut.ut_name[0])		/* login */
	{
	    for(i = 0; i < MAXTTY; i++)
		if (ttyname[i][0] == NUL)
		{
		    strncpy(&ttyname[i][0], ut.ut_line, sizeof(ut.ut_line));
		    ttyname[i][sizeof(ut.ut_line) + 1] = NUL;
		    break;
		}
		else if (!strncmp(&ttyname[i][0], ut.ut_line,
		    					sizeof(ut.ut_line)))
		    break;
		
	    if (i < MAXTTY)
	    {
	    	if (avail[i])
		    ntty ++;
	    	avail[i] = 0;
		if (trace)
		    printf(" ");
	    }
	    else if (trace & TRACE_CTRL)
		printf(" ");
	}
	else				/* logout */
	{
	    for(i = 0; i < MAXTTY; i++)
		if (!strncmp(&ttyname[i][0], ut.ut_line, sizeof(ut.ut_line)))
		    break;

	    if (i < MAXTTY)
	    {
	    	if (!avail[i])
		    ntty --;
		avail[i] = 1;

		if (trace)
		    printf(ntty == 0 ? ">": "."); 
		if (ntty == 0)
		    pos = cpos;
	    }
	    else if (trace)
		printf(".");
	}

	if (trace)
	{
	    tmp = localtime(&ut.ut_time);
	    printf(" %2d/%2d %2d:%02d(%2d) ",	tmp->tm_mon + 1, tmp->tm_mday,
 	   					tmp->tm_hour, tmp->tm_min,
						ntty);

	    for(i = 0; i < MAXTTY; i++)
		if (ttyname[i][0])
		    printf("%c", avail[i]? ' ': '*');

	    printf("\n");
	}
    }

    /*
     *	copy wtmp into new wtmp
     */
    lseek(fut, pos, SEEK_SET); 	/* set current posission */

    fnew = open(newfile, O_WRONLY | O_CREAT, S_IWUSR | S_IRUSR|S_IRGRP|S_IROTH);
    if (fnew < 0)
    {
    	fprintf(stderr, "file %s cannot open\n", newfile);
	return;
    }

    LOCKFILE(fut);
    while((i = read(fut, utbuf, sizeof(utbuf))) > 0)
    	write(fnew, utbuf, i);

    close(fnew);
    UNLKFILE(fut);
    close(fut);
    rename(newfile, file);
}

